use std::convert::TryInto;

impl DNSHeader {
    const QR_MASK: u8 = 0b10000000;
    const OPCODE_MASK: u8 = 0b01111000;
    const AA_MASK: u8 = 0b00000100;
    const TC_MASK: u8 = 0b00000010;
    const RD_MASK: u8 = 0b00000001;

    const RA_MASK: u8 = 0b10000000;
    // no support of rfc4035 at the moment
    //const AD_MASK: u8 = 0b00100000;
    //const CD_MASK: u8 = 0b00010000;
    const RCODE_MASK: u8 = 0b00001111;

    const OPCODE_OFFSET: u8 = 3;

    pub fn from_udp_datagram(datagram: &[u8]) -> Result<Self, DNSParseError> {
        if datagram.len() < 12 {
            return Err(DNSParseError::DatagramTooShort);
        }

        let id = u16::from_be_bytes((&datagram[..2]).try_into().unwrap());

        let message_type = DNSMessageType::from(datagram[2] & Self::QR_MASK != 0);
        let opcode = DNSOpCode::try_from((datagram[2] & Self::OPCODE_MASK) >> Self::OPCODE_OFFSET)?;

        let authorative_answer = (datagram[2] & Self::AA_MASK) != 0;
        let truncated = (datagram[2] & Self::TC_MASK) != 0;
        let recursion_desired = (datagram[2] & Self::RD_MASK) != 0;
        let recursion_available = (datagram[3] & Self::RA_MASK) != 0;

        // no support for rfc4035 at the moment
        //let authentic_data = (datagram[2] & Self::AD_MASK) != 0;
        //let checking_disabled = (datagram[2] & Self::CD_MASK) != 0;

        let response_code = DNSRCode::try_from(datagram[3] & Self::RCODE_MASK)?;
        let query_count = u16::from_be_bytes((datagram[4..6]).try_into().unwrap());
        let answer_count = u16::from_be_bytes((datagram[6..8]).try_into().unwrap());
        let name_server_count = u16::from_be_bytes((datagram[8..10]).try_into().unwrap());
        let additional_count = u16::from_be_bytes((datagram[10..12]).try_into().unwrap());

        Ok(DNSHeader {
            id,
            message_type,
            opcode,
            authorative_answer,
            truncated,
            recursion_desired,
            recursion_available,
            //authentic_data,    // no support of rfc4035 at the moment
            //checking_disabled, // no support of rfc4035 at the moment
            response_code,
            query_count,
            answer_count,
            name_server_count,
            additional_count,
        })
    }
}

#[derive(Debug, Clone, PartialEq)]
pub enum DNSParseError {
    DatagramTooShort,
    DNSOpCodeInvalid,
    DNSRCodeInvalid,
}

#[cfg(test)]
mod tests {

    mod from_udp_datagram {

        mod header_length {
            use super::super::super::*;

            #[test]
            fn too_short() {
                // minimal header with 1 byte missing
                let dns_query = [
                    0x13, 0x37, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result.err().unwrap(), DNSParseError::DatagramTooShort);
            }
        }

        mod opcode {
            use super::super::super::*;

            #[test]
            fn invalid() {
                let mut dns_query = [
                    0xff, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let invalid_opcodes = [3, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15];

                for opcode in invalid_opcodes {
                    dns_query[2] = opcode << 3;
                    let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                    assert_eq!(
                        parse_result.err(),
                        Some(DNSParseError::DNSOpCodeInvalid),
                        "query: {:02x?}, opcode: {}",
                        dns_query,
                        opcode
                    );
                }
            }

            #[test]
            fn valid() -> Result<(), DNSParseError> {
                let mut dns_query = [
                    0xff, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let valid_opcodes = [
                    (0, DNSOpCode::Query),
                    (1, DNSOpCode::IQuery),
                    (2, DNSOpCode::Status),
                    (4, DNSOpCode::Notify),
                    (5, DNSOpCode::Update),
                ];

                for (opcode, parsed_opcode) in valid_opcodes {
                    dns_query[2] = opcode << 3;
                    let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                    assert_eq!(
                        parse_result?.opcode, parsed_opcode,
                        "query: {:02x?}, opcode: {}",
                        dns_query, opcode
                    );
                }
                Ok(())
            }
        }

        mod message_type {
            use super::super::super::*;

            #[test]
            fn query() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.message_type, DNSMessageType::Query);
                Ok(())
            }

            #[test]
            fn response() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.message_type, DNSMessageType::Response);
                Ok(())
            }
        }

        mod response_code {
            use super::super::super::*;

            #[test]
            fn invalid() {
                let mut dns_query = [
                    0xff, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let invalid_rcodes = [11, 12, 13, 14, 15];

                for rcode in invalid_rcodes {
                    dns_query[3] = rcode;

                    let parse_result = DNSHeader::from_udp_datagram(&dns_query);

                    assert_eq!(
                        parse_result,
                        Err(DNSParseError::DNSRCodeInvalid),
                        "query: {:02x?}, rcode: {}",
                        dns_query,
                        rcode,
                    );
                }
            }

            #[test]
            fn valid() -> Result<(), DNSParseError> {
                let mut dns_query = [
                    0xff, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];

                // for now we only support RCodes 0 to 10 (inclusive)
                let valid_rcodes = [
                    (0, DNSRCode::NoError),
                    (1, DNSRCode::FormErr),
                    (2, DNSRCode::ServFail),
                    (3, DNSRCode::NXDomain),
                    (4, DNSRCode::NotImp),
                    (5, DNSRCode::Refused),
                    (6, DNSRCode::YXDomain),
                    (7, DNSRCode::YXRRSet),
                    (8, DNSRCode::NXRRSet),
                    (9, DNSRCode::NotAuth),
                    (10, DNSRCode::NotZone),
                ];

                for (rcode, parsed_rcode) in valid_rcodes {
                    dns_query[3] = rcode;
                    let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                    assert_eq!(
                        parse_result?.response_code, parsed_rcode,
                        "query: {:02x?}, rcode: {}, parsed_rcode: {:?}",
                        dns_query, rcode, parsed_rcode
                    );
                }
                Ok(())
            }
        }

        mod query_count {
            use super::super::super::*;

            #[test]
            fn zero() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.query_count, 0);
                Ok(())
            }

            #[test]
            fn one() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x80, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.query_count, 1);
                Ok(())
            }

            #[test]
            fn max_value() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x80, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.query_count, 65535);
                Ok(())
            }
        }

        mod answer_count {
            use super::super::super::*;

            #[test]
            fn zero() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.answer_count, 0);
                Ok(())
            }

            #[test]
            fn one() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.answer_count, 1);
                Ok(())
            }

            #[test]
            fn max_value() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x80, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.answer_count, 65535);
                Ok(())
            }
        }

        mod name_server_count {
            use super::super::super::*;

            #[test]
            fn zero() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.name_server_count, 0);
                Ok(())
            }

            #[test]
            fn one() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.name_server_count, 1);
                Ok(())
            }

            #[test]
            fn max_value() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.name_server_count, 65535);
                Ok(())
            }
        }

        mod additional_count {
            use super::super::super::*;

            #[test]
            fn zero() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.additional_count, 0);
                Ok(())
            }

            #[test]
            fn one() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.additional_count, 1);
                Ok(())
            }

            #[test]
            fn max_value() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert_eq!(parse_result?.additional_count, 65535);
                Ok(())
            }
        }

        mod authorative_answer {
            use super::super::super::*;

            #[test]
            fn yes() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert!(parse_result?.authorative_answer);
                Ok(())
            }

            #[test]
            fn no() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert!(!parse_result?.authorative_answer);
                Ok(())
            }
        }

        mod truncated {
            use super::super::super::*;

            #[test]
            fn yes() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert!(parse_result?.truncated);
                Ok(())
            }

            #[test]
            fn no() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert!(!parse_result?.truncated);
                Ok(())
            }
        }

        mod recursion_desired {
            use super::super::super::*;

            #[test]
            fn yes() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert!(parse_result?.recursion_desired);
                Ok(())
            }

            #[test]
            fn no() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert!(!parse_result?.recursion_desired);
                Ok(())
            }
        }

        mod recursion_available {
            use super::super::super::*;

            #[test]
            fn yes() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert!(parse_result?.recursion_available);
                Ok(())
            }

            #[test]
            fn no() -> Result<(), DNSParseError> {
                let dns_query = [
                    0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                ];
                let parse_result = DNSHeader::from_udp_datagram(&dns_query);
                assert!(!parse_result?.recursion_available);
                Ok(())
            }
        }
    }
}
